from os import environ

from typer import Typer

from kolombo.bin.domain import cli as domain_cli
from kolombo.bin.run import cli as run_cli
from kolombo.bin.user import cli as user_cli
from kolombo.bin.util import CliLog, run_command

cli = Typer()
cli.add_typer(user_cli, name="user")
cli.add_typer(domain_cli, name="domain")
cli.add_typer(run_cli, name="run")
#: Default content of /etc/kolombo/kolombo.env file
_default_env = """# This is configuration file for kolombo that is used by default
# To enable debug - change to 1 here
DEBUG=0
# Secret key that is used to determine that nginx is using API
# NB! Change this secret!
NGINX_SECRET_KEY=changeme
# Maximum auth attempts
MAX_ATTEMPTS=3
# Salt used for passwords hashing
# NB! Change this secret!
SALT=changeme
# Name of SQLite database file
DATABASE_NAME=kolombo
"""
#: Default content of /etc/kolombo/virtual/ files
_virtual = {
    "addresses": "bob@example.com bob@example.com",
    "domains": "example.com",
    "mailbox": "bob@example.com bob@example.com/",
    "ssl_map": (
        "example.com /etc/letsencrypt/live/example.com/privkey.pem "
        "/etc/letsencrypt/live/example.com/fullchain.pem"
    ),
}


@cli.command("setup")
def setup() -> None:
    """Set up kolombo for current user"""
    log = CliLog()
    log.step("- Installing bash completion")
    run_command("kolombo --install-completion bash")

    log.step("- Creating /etc/kolombo folders and files")
    log.step("-- Creating /etc/kolombo folder (using sudo)")
    run_command("mkdir -p -m 770 /etc/kolombo", sudo=True)

    user = environ["USER"]
    log.step(f"-- Setting /etc/kolombo owner to {user} (using sudo)")
    run_command(f"chown {user}:{user} /etc/kolombo", sudo=True)

    log.step("-- Creating /etc/kolombo folders for volumes")
    run_command("mkdir -p -m 770 /etc/kolombo/maildirs")
    run_command("mkdir -p -m 770 /etc/kolombo/mail-enabled")
    run_command("mkdir -p -m 770 /etc/kolombo/virtual")
    run_command("mkdir -p -m 700 /etc/kolombo/dkim_keys")

    log.step("-- Writing default conf to /etc/kolombo/kolombo.env")
    with open("/etc/kolombo/kolombo.env", "w") as default_config:
        default_config.write(_default_env)

    for file in ("addresses", "domains", "mailbox", "ssl_map"):
        log.step(f"-- Writing default file to /etc/kolombo/virtual/{file}")
        with open(f"/etc/kolombo/virtual/{file}", "w") as virtual_file:
            virtual_file.write(f"{_virtual[file]}\n")


def main() -> None:
    cli()



from asyncio import run
from functools import wraps
from importlib.resources import path as resource_path

# Security implications are considered
from subprocess import STDOUT, CalledProcessError, check_output  # nosec: B404
from typing import Any, Awaitable, Callable, List, Optional, Union

from rich.console import Console
from rich.markdown import Markdown


def async_command(func: Callable[..., Awaitable[None]]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return run(func(*args, **kwargs))

    return wrapper


def run_command(command: str, sudo: bool = False) -> str:
    cmd = command.split(" ")
    if sudo:
        cmd.insert(0, "sudo")
    # We trust input here, it does not come from user
    return check_output(cmd, text=True, stderr=STDOUT)  # nosec: B603


def docker(command: str) -> str:
    return run_command(f"docker {command}", sudo=True)


def kolombo_image_exists(component: str) -> bool:
    try:
        docker(f"image inspect kolombo-{component}")
        return True
    except CalledProcessError:
        return False


def build_kolombo_image(component: str) -> str:
    with resource_path(package="kolombo", resource=".") as context:
        dockerfile = f"{context}/docker/{component}/Dockerfile"
        return docker(
            f"build --no-cache -f {dockerfile} -t kolombo-{component} {context}"
        )


def run_container(
    image: str,
    args: Optional[str] = None,
    *,
    name: Optional[str] = None,
    net: Optional[str] = None,
    volumes: Optional[List[str]] = None,
    ports: Optional[List[str]] = None,
    other_flags: str = "-d",
) -> str:
    cmd = "run"
    if args is not None:
        image += f" {args}"
    if name is not None:
        cmd += f" --name {name}"
    if net is not None:
        cmd += f" --net {net}"
    if volumes:
        cmd += f" -v {' -v '.join(volumes)}"
    if ports:
        cmd += f" -p {' -p '.join(ports)}"

    return docker(f"{cmd} {other_flags} {image}")


def create_network(network: str, subnet: str) -> str:
    # Create only if does not exist
    if not docker(f"network ls -q --filter name={network}"):
        return docker(f"network create {network} --subnet={subnet}")
    return "Network already created"


class CliLog:
    def __init__(self) -> None:
        self._console = Console()

    def step(self, message: Union[str, Markdown]) -> None:
        self._console.print(message, style="bold green")

    def info(self, message: Union[str, Markdown]) -> None:
        self._console.print(message, style="bold cyan")

    def error(self, message: Union[str, Markdown]) -> None:
        self._console.print(message, style="bold red")



from rich.markdown import Markdown
from typer import Argument, Option, Typer

from kolombo import conf
from kolombo.bin.util import (
    CliLog,
    async_command,
    build_kolombo_image,
    create_network,
    kolombo_image_exists,
    run_container,
)
from kolombo.models import Domain
from kolombo.resources import init_database

cli = Typer()


@cli.command("list")
@async_command
async def list_domains(
    conf: str = Option(None, help="Path to .env file with configuration")  # noqa: B008
) -> None:
    log = CliLog()
    await init_database()

    all_domains = await Domain.all_active()
    active_pairs = [f"{domain.mx} -> {domain.actual}" for domain in all_domains]
    log.step(f"Active domains: {len(active_pairs)}")
    if len(active_pairs) > 0:
        log.info(Markdown(f"*  {'* '.join(active_pairs)}"))


async def _update_virtual_domains() -> None:
    domains_list = "\n".join(domain.actual for domain in await Domain.all_active())
    with open("/etc/kolombo/virtual/domains", mode="w") as domains_file:
        domains_file.write(f"# File is auto-generated\n{domains_list}\n")


async def _update_virtual_ssl_map() -> None:
    domains_mx = [domain.mx for domain in await Domain.all_active()]
    ssl_map = "\n".join(
        f"{mx} /etc/letsencrypt/live/{mx}/privkey.pem "
        f"/etc/letsencrypt/live/{mx}/fullchain.pem"
        for mx in domains_mx
    )
    with open("/etc/kolombo/virtual/ssl_map", mode="w") as ssl_map_file:
        ssl_map_file.write(f"{ssl_map}\n")


_nginx_config = """# File was auto-generated by kolombo
# Receive-only SMTP & IMAP mail reading
server {
    server_name {{MX_DOMAIN}};
    ssl_certificate /etc/letsencrypt/live/{{MX_DOMAIN}}/fullchain.pem;
    ssl_certificate_key /etc/letsencrypt/live/{{MX_DOMAIN}}/privkey.pem;
    auth_http_header X-Secret-Key {{SECRET_KEY}};
    auth_http_header X-Domain {{ACTUAL_DOMAIN}};
    auth_http kolombo-auth:7089/auth;

    listen 993 ssl;
    protocol imap;
}

# Send-only SMTP
server {
    server_name {{MX_DOMAIN}};
    ssl_certificate /etc/letsencrypt/live/{{MX_DOMAIN}}/fullchain.pem;
    ssl_certificate_key /etc/letsencrypt/live/{{MX_DOMAIN}}/privkey.pem;
    auth_http_header X-Secret-Key {{SECRET_KEY}};
    auth_http_header X-Domain {{ACTUAL_DOMAIN}};
    auth_http kolombo-auth:7089/auth;

    listen 587 ssl;
    protocol smtp;
}
"""


def _add_mail_enabled_config(domain: str, mx: str) -> None:
    config = _nginx_config.replace("{{ACTUAL_DOMAIN}}", domain)
    config = config.replace("{{MX_DOMAIN}}", mx)
    config = config.replace("{{SECRET_KEY}}", conf.NGINX_SECRET_KEY)
    with open(f"/etc/kolombo/mail-enabled/{mx}.conf", mode="w") as nginx_file:
        nginx_file.write(config)


def _generate_dkim_key(domain: str) -> str:
    create_network("kolombo", subnet="192.168.79.0/24")
    return run_container(
        "kolombo-sender",
        args=f"gen_key {domain}",
        name=f"kolombo-{domain}-sender",
        net="kolombo",
        volumes=["/etc/kolombo/dkim_keys:/etc/opendkim/keys"],
        # Remove container after stop, attach interactive TTY
        other_flags="--rm -it",
    )


def _parse_txt_record(output: str) -> str:
    txt_record = output[output.find("(") + 1 : output.find(")")]  # noqa: E203
    return txt_record.replace('"\n\t  "', "")


@cli.command("add")
@async_command
async def add_domain(
    conf: str = Option(None, help="Path to .env file with configuration"),  # noqa: B008
    domain: str = Argument(..., help="Domain that come after @ in email"),  # noqa: B008
    mx: str = Argument(None, help="Domain from DNS MX record if exists"),  # noqa: B008
) -> None:
    log = CliLog()
    await init_database()

    if mx is None:
        mx = domain

    if not domain or not mx:
        log.error("Arguments MUST NOT be empty string")
        exit(1)
    elif await Domain.objects.filter(actual=domain, mx=mx).exists():
        log.error(f"Domain pair '{mx} -> {domain}' exists")
        exit(2)

    log.step("- Adding config to mail-enabled")
    _add_mail_enabled_config(domain, mx)
    # TODO: Do postfix reload in kolombo-receiver
    log.step("- Building kolombo-sender for DKIM key generation (if does not exist)")
    if not kolombo_image_exists(component="sender"):
        build_kolombo_image(component="sender")

    log.step("- Generating DKIM key ([u]save TXT DNS record from output[/u])")
    output = _generate_dkim_key(domain)
    log.info(f"mail._domainkey.{domain} TXT: {_parse_txt_record(output)}")
    log.step(f"- Adding pair '{mx} -> {domain}' to database")
    await Domain.objects.create(actual=domain, mx=mx)
    log.step("- Updating virtual domains")
    await _update_virtual_domains()
    await _update_virtual_ssl_map()
    log.step(f"Domain pair '{mx} -> {domain}' added!")



import os
import re

from setuptools import find_packages, setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    with open(os.path.join(package, "__init__.py")) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


def get_long_description():
    """
    Return the README.
    """
    with open("README.md", encoding="utf8") as f:
        return f.read()


setup(
    name="kolombo",
    python_requires=">=3.7",
    version=get_version("kolombo"),
    description="Kolombo - easy to manage mail server 💌",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github/HarrySky/kolombo",
    license="Apache License 2.0",
    author="Igor Nehoroshev",
    author_email="hi@neigor.me",
    maintainer="Igor Nehoroshev",
    maintainer_email="hi@neigor.me",
    packages=find_packages(),
    # Use MANIFEST.in for data files
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        # Fast async event loop
        "uvloop==0.14.0",
        # For CLI
        "typer==0.3.2",
        "rich==9.3.0",
        # ORM
        "ormar==0.7.1",
        "databases==0.4.1",
        "aiosqlite==0.16.0",
        # Passwords hashing
        "cryptography==3.2.1",
        # For API
        "fastapi==0.62.0",
        "uvicorn==0.12.3",
        # Since 0.12.0 uvicorn does not install httptools (and uvloop)
        "httptools==0.1.*",
    ],
    entry_points={"console_scripts": ["kolombo = kolombo:cli"]},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications :: Email",
        "Typing :: Typed",
    ],
)
